import logging

from ..sdk import VerseSDK


def apply_integrations(sdk: VerseSDK, vendor: str):
    try:
        if vendor == "pydantic_ai":
            from .pydantic_ai import patch_pydantic_ai

            patch_pydantic_ai(sdk)
            logging.info("`pydantic_ai` integration applied to `VerseSDK`")
        elif vendor == "litellm":
            from .litellm import configure_litellm

            configure_litellm(sdk)
            logging.info("`litellm` integration applied to `VerseSDK`")
    except Exception as e:
        logging.warning(f"Error applying `{vendor}` integration", exc_info=e)


__all__ = ["apply_integrations"]



import logging
from typing import Any

from ...contexts import CompletionChoice, GenerationUsage


def calculate_total_tokens(
    prompt_tokens: int | None, completion_tokens: int | None
) -> int | None:
    """Calculate the total tokens."""
    if prompt_tokens is not None or completion_tokens is not None:
        return (prompt_tokens or 0) + (completion_tokens or 0)
    return None


def get_completition_choices(data: Any) -> list[CompletionChoice]:
    """Get the completion choices from the response."""

    choices = []
    if hasattr(data, "choices"):
        choices = data.choices
    elif isinstance(data, dict) and "choices" in data:
        choices = data["choices"]

    return choices if len(choices) > 0 else []


def get_finish_reason(response_obj: Any) -> str | None:
    """Trace response metadata like model and ID."""
    if getattr(response_obj, "choices", None):
        fr = getattr(response_obj.choices[0], "finish_reason", None)
        if fr:
            return fr

    return None


def get_operation_type(messages: list, kwargs: dict) -> str:
    try:
        """Determine the operation type based on request parameters."""
        if kwargs.get("litellm_params", {}).get("api_base", "").endswith("/embeddings"):
            return "embedding"

        return "chat" if messages else "completion"
    except Exception as e:
        logging.warning("Error determining operation type from Litellm", exc_info=e)
        return "completion"


def get_stream_text(response_obj: Any) -> str:
    """Get the text from the response."""
    chunk = getattr(response_obj, "choices", None)
    if not chunk:
        return ""

    c = chunk[0]
    return (
        getattr(getattr(c, "delta", None), "content", None)
        or getattr(c, "text", None)
        or ""
    )


def get_usage(kwargs: dict) -> GenerationUsage:
    """Get the usage for the request."""
    try:
        optional = kwargs.get("optional_params", {})
        return {
            "max_tokens": kwargs.get("max_tokens", optional.get("max_tokens")),
            "temperature": kwargs.get("temperature", optional.get("temperature")),
            "top_p": kwargs.get("top_p", optional.get("top_p")),
            "stream": kwargs.get("stream", optional.get("stream")),
        }
    except Exception as e:
        logging.warning("Error getting usage from Litellm", exc_info=e)
        return {}


def get_usage_from_response(response_obj: Any) -> GenerationUsage:
    """Get the usage from the response."""
    try:
        usage = getattr(response_obj, "usage", None)

        usage_output = {
            "prompt_tokens": getattr(usage, "prompt_tokens", None),
            "completion_tokens": getattr(usage, "completion_tokens", None),
        }

        usage_output["total_tokens"] = getattr(
            usage,
            "total_tokens",
            calculate_total_tokens(
                usage_output["prompt_tokens"],
                usage_output["completion_tokens"],
            ),
        )

        return usage_output

    except Exception as e:
        logging.warning("Error getting usage from response", exc_info=e)
        return {}



import pytest

from tests.fixtures.agents import AgentFixtures
from verse_sdk import verse


@pytest.mark.e2e
@pytest.mark.asyncio
async def test_pydantic_ai_langfuse_http_export():
    with verse.trace("pydantic_ai_e2e") as t:
        agent = AgentFixtures().create_pydantic_agent()
        resp = await agent.ask_about_friends("Ollie the Owl")
        t.update(output=resp)



import json

from verse_sdk import verse

verse.init(
    app_name="test_context_managers",
    exporters=[verse.exporters.console()],
)


def parse_spans(out: str) -> list[dict]:
    """Parse the spans from the output."""
    spans = []
    buffer = ""
    for line in out.splitlines():
        line = line.strip()
        if not line:
            continue
        buffer += line
        if buffer.count("{") == buffer.count("}"):
            try:
                span = json.loads(buffer)
                spans.append(span)
            except json.JSONDecodeError:
                pass
            buffer = ""

    spans.reverse()
    return spans


def test_context_managers(capfd):
    with verse.trace(name="test_trace") as trace:
        trace.input("test_input")
        trace.output("test_output")

        with verse.span(name="test_span") as span:
            span.input("test_span_input")
            span.output("test_span_output")

            with verse.generation(name="test_generation") as generation:
                generation.model("test_model")
                generation.input("test_generation_input")
                generation.output("test_generation_output")

    out, _ = capfd.readouterr()
    spans = parse_spans(out)
    assert len(spans) == 3

    assert spans[0]["name"] == "test_trace"
    assert spans[0]["attributes"]["input"] == "test_input"
    assert spans[0]["attributes"]["output"] == "test_output"
    trace_id = spans[0]["context"]["trace_id"]

    assert spans[1]["name"] == "test_span"
    assert spans[1]["context"]["trace_id"] == trace_id
    assert spans[1]["attributes"]["input"] == "test_span_input"
    assert spans[1]["attributes"]["output"] == "test_span_output"

    assert spans[2]["name"] == "test_generation"
    assert spans[2]["context"]["trace_id"] == trace_id
    assert spans[2]["attributes"]["gen_ai.request.model"] == "test_model"
    assert spans[2]["attributes"]["gen_ai.prompt"] == "test_generation_input"
    assert spans[2]["attributes"]["gen_ai.completion"] == "test_generation_output"

def test_session_and_project_propagation(capfd):
    """Test that session and project IDs are propagated from parent to child spans."""
    with verse.trace(
        name="parent_trace", session_id="test-session-123", project_id="test-project-456"
    ) as trace:
        trace.input("parent_input")

        with verse.span(name="child_span") as span:
            span.input("child_span_input")

            with verse.generation(name="child_generation") as generation:
                generation.model("test_model")
                generation.input("child_generation_input")

    out, _ = capfd.readouterr()
    spans = parse_spans(out)
    assert len(spans) == 3

    assert spans[0]["name"] == "parent_trace"
    assert spans[0]["attributes"]["session.id"] == "test-session-123"
    assert spans[0]["attributes"]["project.id"] == "test-project-456"

    assert spans[1]["name"] == "child_span"
    assert spans[1]["attributes"]["session.id"] == "test-session-123"
    assert spans[1]["attributes"]["project.id"] == "test-project-456"

    assert spans[2]["name"] == "child_generation"
    assert spans[2]["attributes"]["session.id"] == "test-session-123"
    assert spans[2]["attributes"]["project.id"] == "test-project-456"



from dotenv import load_dotenv

from verse_sdk import verse

load_dotenv()


verse.init(
    app_name="verse_langfuse_litellm_e2e",
    environment="development",
    exporters=[verse.exporters.langfuse()],
    vendor="litellm",
    version="1.0.0",
)



from typing import Any, ClassVar

from opentelemetry.trace import Status, StatusCode

from ...contexts import GenerationContext, GenerationMessage
from ...sdk import VerseSDK
from ...utils import apply_value
from .utils import (
    get_completition_choices,
    get_finish_reason,
    get_operation_type,
    get_stream_text,
    get_usage,
    get_usage_from_response,
)


def configure_litellm(sdk: VerseSDK):
    """Configure Litellm to use the Verse SDK."""
    import litellm
    from litellm.integrations.custom_logger import CustomLogger

    class LitellmCallbackHandlers(CustomLogger):
        current_spans: ClassVar[dict[str, GenerationContext]] = {}

        async def async_log_failure_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle failed async LLM requests"""
            self.log_failure_event(kwargs, response_obj, start_time, end_time)

        async def async_log_stream_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Called for each streaming chunk."""
            self.log_stream_event(kwargs, response_obj, start_time, end_time)

        async def async_log_success_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle successful async LLM requests"""
            self.log_success_event(kwargs, response_obj, start_time, end_time)

        def log_failure_event(
            self, kwargs, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle failed LLM requests"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                span_ctx.error(kwargs.get("exception", response_obj))
                self._drop_current_span(kwargs)

        def log_post_api_call(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """End the generation span and remove from the active context variable"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                sdk.event(
                    "llm.response_received",
                    metadata={"length": len(str(response_obj))},
                    level="info",
                    span=span_ctx.span,
                )

        def log_pre_api_call(
            self, model: str, messages: list[GenerationMessage], kwargs: dict
        ):
            """Create a generation span around the LLM request"""
            id = kwargs.get("litellm_call_id")
            self.current_spans[id] = sdk.create_generation(
                name=f"LiteLLM: {(get_operation_type(messages, kwargs)).capitalize()}",
                messages=messages,
                metadata=kwargs.get("metadata", {}),
                model=model,
                usage=get_usage(kwargs),
            )

        def log_stream_event(
            self, kwargs: dict, response_obj: Any, start_time: float, end_time: float
        ):
            """Called for each streaming chunk."""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                try:
                    text = get_stream_text(response_obj)
                    if text:
                        span_ctx.event(
                            "gen_ai.stream.delta",
                            metadata={"content": text[:512]},
                        )
                except Exception:
                    pass

        def log_success_event(
            self, kwargs, response_obj: Any, start_time: float, end_time: float
        ):
            """Handle successful LLM requests"""
            span_ctx = self._get_current_span(kwargs)

            if span_ctx:
                try:
                    apply_value(span_ctx, "model_used", response_obj.model)
                    apply_value(span_ctx, "reason", get_finish_reason(response_obj))

                    span_ctx.completions(
                        get_completition_choices(
                            kwargs.get("complete_streaming_response", response_obj)
                        )
                    )

                    span_ctx.usage(get_usage_from_response(response_obj))
                    span_ctx.status(Status(StatusCode.OK))
                except Exception as e:
                    span_ctx.error(e)
                finally:
                    self._drop_current_span(kwargs)

        def _drop_current_span(self, kwargs: dict):
            span = self.current_spans.pop(kwargs.get("litellm_call_id"), None)
            if span:
                span.span.end()

        def _get_current_span(self, kwargs: dict) -> GenerationContext:
            return self.current_spans.get(kwargs.get("litellm_call_id"))

    litellm.callbacks = [LitellmCallbackHandlers()]



from __future__ import annotations

import logging
from typing import Any

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.trace import ProxyTracerProvider

from .decorators import with_decorators
from .exporters import (
    Exporter,
    ExportersRegistry,
)
from .integrations import apply_integrations
from .processors import AttributeProcessor
from .sdk import VerseSDK


class VerseSDKBuilder:
    exporters = ExportersRegistry()

    _initialized: bool = False
    _sdk: VerseSDK | None
    _tracer: trace.Tracer | None

    def __getattr__(self, name: str) -> Any:
        """Dynamically access methods from the SDK."""
        if not self._sdk:
            raise ValueError("SDK not initialized")

        return getattr(self._sdk, name)

    def flush(self, timeout_ms: int = 30000) -> None:
        """
        Flush all pending traces to exporters.

        Args:
            timeout_ms: Maximum time to wait for flush in milliseconds
        """
        try:
            if not self._initialized:
                return

            provider = trace.get_tracer_provider()
            if isinstance(provider, ProxyTracerProvider):
                provider.force_flush(timeout_millis=timeout_ms)
        except Exception as e:
            logging.warning("Error flushing traces", exc_info=e)

    def init(
        self,
        app_name: str = "verse_sdk_observability",
        environment: str | None = "development",
        exporters: list[Exporter] | None = None,
        vendor: str | None = None,
        version: str | None = "1.0.0",
    ) -> VerseSDK:
        """
        Lazily initialize the observability client.

        Args:
            app_name: Service name for tracing
            exporters: Exporter name(s) or None for auto-detection
            environment: Deployment environment (e.g., "production", "staging")
            version: Service version

        Returns:
            VerseSDK w/ decorations
        """
        try:
            if self._initialized:
                return self._sdk

            resource = Resource.create(
                {
                    "deployment.environment": environment,
                    "service.name": app_name,
                    "service.version": version,
                }
            )

            provider = TracerProvider(resource=resource)

            provider.add_span_processor(AttributeProcessor())

            for exporter in exporters:
                try:
                    provider.add_span_processor(
                        exporter.create_span_processor(resource)
                    )
                    logging.info(
                        "Added span processor `%s`",
                        exporter.__class__.__name__,
                    )
                except Exception as e:
                    logging.warning(
                        "Error adding span processor `%s`",
                        exporter.__class__.__name__,
                        exc_info=e,
                    )

            trace.set_tracer_provider(provider)

            self._tracer = trace.get_tracer(app_name)
            self._sdk = VerseSDK(self._tracer)
            self._sdk = with_decorators(self._sdk)
            self._initialized = True

            apply_integrations(self._sdk, vendor)
            logging.info("🚀 VerseSDK initialized")

            return self._sdk
        except Exception as e:
            logging.error("Error initializing SDK", exc_info=e)
            raise e



import pytest

from tests.fixtures.agents import AgentFixtures
from verse_sdk import verse


@pytest.mark.e2e
def test_litellm_langfuse_http_export():
    with verse.trace("litellm_e2e") as t:
        agent = AgentFixtures().create_litellm_agent()
        resp = agent.ask_about_friends("Ollie the Owl")
        t.update(output=resp)


@pytest.mark.e2e
@pytest.mark.asyncio
async def test_litellm_langfuse_http_export_async():
    with verse.trace("litellm_e2e"):
        agent = AgentFixtures().create_litellm_agent()
        async for _ in agent.tell_story_about_character("Cleo the Cat"):
            pass



import functools
import logging
from collections.abc import AsyncIterable

from ...sdk import VerseSDK


def patch_pydantic_ai(sdk: VerseSDK):
    """Patch Pydantic AI to use the Verse SDK."""

    from pydantic_ai import Agent, RunContext
    from pydantic_ai.messages import (
        AgentStreamEvent,
        PartDeltaEvent,
    )

    from .utils import (
        get_finish_reason,
        get_generation_messages,
        get_llm_info_from_model,
        get_operation_type,
        get_usage,
    )

    async def handle_events(ctx: RunContext, events: AsyncIterable[AgentStreamEvent]):
        """
        Handle Pydantic events to create LLM generation spans.
        Pydantic dispatches more events than we need, so the delta and finish reason checks are necessary to avoid duplicates.
        """
        try:
            contains_generation = get_finish_reason(ctx) == "stop"

            async for event in events:
                if isinstance(event, PartDeltaEvent):
                    contains_generation = True

            if not contains_generation:
                return

            op = get_operation_type(ctx)
            vendor, model_name = get_llm_info_from_model(ctx.model)

            sdk.create_generation(
                f"Pydantic AI: {op.capitalize()}",
                messages=get_generation_messages(ctx),
                model=model_name,
                usage=get_usage(ctx.usage),
                vendor=vendor,
            ).span.end()
        except Exception as e:
            logging.warning("Pydantic streaming event dropped", exc_info=e)

    def patch():
        """Setup monkey-patches for Pydantic AI if not already patched."""
        if not getattr(Agent, "_verse_patched", False):
            patch_run()
            Agent._verse_patched = True

    def patch_run():
        """Setup a moneky patch for Agent.run() that captures the LLM call in a generation span."""
        original_run_method = Agent.run

        @functools.wraps(original_run_method)
        async def run__patched(self, *args, **kwargs):
            try:
                return await original_run_method(
                    self, *args, **kwargs, event_stream_handler=handle_events
                )
            except Exception as e:
                logging.warning("Error patching Agent.run()", exc_info=e)
                return original_run_method(self, *args, **kwargs)

        Agent.run = run__patched

    patch()



from tests.fixtures.agents.litellm_agent import LitellmAgent
from tests.fixtures.agents.pydantic_agent import PydanticAgent


class AgentFixtures:
    """Factory for different agent fixtures"""

    def create_litellm_agent(self) -> LitellmAgent:
        """Create a Litellm agent"""
        return LitellmAgent()

    def create_pydantic_agent(self) -> PydanticAgent:
        """Create a Pydantic agent"""
        return PydanticAgent()



from tests.fixtures.agents.prompt import SYSTEM_PROMPT
from tests.fixtures.datasource import datasource
from verse_sdk import verse

from .litellm import get_completion_text, run_completion, run_completion_async


class LitellmAgent:
    def ask_about_friends(self, character_name: str):
        """
        A sample question that should trigger:
        -   `get_all_characters`
        -   `get_friends`
        """
        input = f"""What are {character_name}'s friends?"""

        with verse.span(
            "Character Friends Agent",
            op="agent",
            input=input,
            scope="test_agent",
        ) as span:
            response = run_completion(
                [
                    {"role": "system", "content": SYSTEM_PROMPT},
                    {"role": "user", "content": input},
                ],
                datasource,
            )

            output = get_completion_text(response)
            span.output(output=output)
            return output

    async def tell_story_about_character(self, character_name: str):
        """
        A sample question that should trigger:
        -   `get_character_by_id`
        -   `get_character_location`
        -   `get_character_occupation`
        """
        input = f"""Tell me a story about the day in the life of {character_name}. Include the following details:
        -   What they do for friends
        -   What they do for work
        -   Where they live
        """

        with verse.span(
            "Character Story Agent",
            op="agent",
            input=input,
            scope="test_agent",
        ) as span:
            output = ""
            parts = run_completion_async(
                [
                    {"role": "system", "content": SYSTEM_PROMPT},
                    {"role": "user", "content": input},
                ],
                datasource,
            )

            async for part in parts:
                output += part
                yield part

            span.output(output=output)



from .decorators import lazily
from .sdk_builder import VerseSDKBuilder

verse = VerseSDKBuilder()
exporters = verse.exporters

observe = lazily(verse, "observe")
observe_generation = lazily(verse, "observe_generation")
observe_span = lazily(verse, "observe_span")
observe_tool = lazily(verse, "observe_tool")
observe_trace = lazily(verse, "observe_trace")


__all__ = [
    "exporters",
    "observe",
    "observe_generation",
    "observe_span",
    "observe_trace",
    "verse",
]



import logging
from functools import wraps
from typing import Any

from .base import Exporter
from .console import ConsoleExporter
from .langfuse import LangfuseExporter
from .langfuse.types import LangfuseConfig
from .otel import OTLPExporter
from .types import ExporterConfig


class ExportersRegistry:
    @staticmethod
    def safe_named_exporter(fn):
        """Create an error boundary around each exporter factory function."""

        @wraps(fn)
        def create_exporter(*args, **kwargs) -> Exporter | None:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                logging.warning("Error creating exporter `%s`", fn.__name__, exc_info=e)
                return None

        return create_exporter

    @safe_named_exporter
    def console(self, config: ExporterConfig = None) -> ConsoleExporter:
        """Create a console exporter."""
        return ConsoleExporter(config or {})

    @safe_named_exporter
    def langfuse(self, config: LangfuseConfig = None) -> LangfuseExporter:
        """Create a langfuse exporter."""
        return LangfuseExporter(config or {})

    @safe_named_exporter
    def otel(self, config: Any = None) -> OTLPExporter:
        """Create a otel exporter."""
        return OTLPExporter(config or {})


__all__ = [
    "Exporter",
    "ExportersRegistry",
]



from pydantic_ai import Agent

from tests.fixtures.agents.prompt import SYSTEM_PROMPT
from tests.fixtures.datasource import datasource
from verse_sdk import verse


class PydanticAgent:
    agent: Agent

    def __init__(self):
        """Initialize the agent"""
        self.agent = Agent(
            model="openai:gpt-4o-mini",
            system_prompt=SYSTEM_PROMPT,
            tools=datasource,
        )

    def ask_about_character(self, character_id: int):
        """
        A sample question that should trigger:
        -   `get_character_by_id`
        -   `get_character_location`
        -   `get_character_occupation`
        """
        return self._ask(
            f"""
            Tell me about character with ID {character_id}.
            Where do they live and what do they do for work?
            """.strip()
        )

    def ask_about_friends(self, character_name: str):
        """
        A sample question that should trigger:
        -   `get_all_characters`
        -   `get_friends`
        """
        return self._ask(f"""What are {character_name}'s friends?""")

    async def _ask(self, query: str):
        """Run a prompt with SDK tracing"""
        with verse.span("Character Agent", op="agent", scope="test_agent") as s:
            response = await self.agent.run(
                query,
            )

            result = response.output
            s.update(input=query, output=result)
            return result



from typing import Any

from ..utils import get
from .types import CompletionChoice, CompletionChoiceMessage, GenerationMessage


def get_choice_message(
    choice: CompletionChoice | None,
) -> CompletionChoiceMessage | None:
    """Get message dict from choice"""
    if hasattr(choice, "message") and choice.message:
        return choice.message
    elif hasattr(choice, "delta") and choice.delta:
        return choice.delta
    elif isinstance(choice, dict):
        return choice.get("message") or choice.get("delta") or None
    return None


def get_content_from_choice_message(message: CompletionChoiceMessage | None) -> str:
    """Get the content from the choice message"""
    content = get(message, "content")

    if content is None:
        content = ""
    elif not isinstance(content, str):
        content = str(content)

    return content


def get_content_from_generation_message(message: GenerationMessage | None) -> str:
    """Get the content from the generation message."""
    content = get(message, "content")

    if isinstance(content, list):
        text_parts = []
        for part in content:
            text = get(part, "text")
            if text:
                text_parts.append(text)
        content = "\n".join(text_parts)
    elif not isinstance(content, str):
        content = str(content)

    return content


def get_function_call(
    message: Any | None,
    default_name: str = "function_call",
) -> tuple[str | None, str | None]:
    """Get function call details from message"""
    func_call = get(message, default_name)
    return (
        (get(func_call, "arguments"), get(func_call, "name"))
        if func_call
        else (None, None)
    )


def get_tool_call_from_choice_message(
    message: CompletionChoiceMessage | None,
) -> list[dict]:
    """Get tools calls from message if available"""
    tool_calls = get(message, "tool_calls", [])
    return tool_calls if isinstance(tool_calls, list) else []


def get_tool_call(
    tool_call: Any | None,
) -> tuple[str | None, str | None]:
    """Get tool call details"""
    return (
        (get(tool_call, "id"), get(tool_call, "type")) if (None, None) else (None, None)
    )



from __future__ import annotations

import inspect
import logging
from functools import wraps

from .contexts import ContextType, OperationType
from .sdk import VerseSDK
from .utils import to_json


def lazily(verse: VerseSDK, decorator_name: str):
    """Create a lazy decorator that evaluates at call time."""

    def decorator(*args, **kwargs):
        try:
            verse_decorator = getattr(verse, decorator_name)
            return verse_decorator(*args, **kwargs)
        except Exception:

            def noop_decorator(fn):
                return fn

            return noop_decorator

    return decorator


def with_decorators(sdk: VerseSDK) -> VerseSDK:
    """A higher-order function for dropping decorators onto the SDK."""

    def create_observation_decorator(
        span_type: ContextType, operation_type: OperationType | None = None
    ):
        def observe_wrapper(
            name: str | None = None,
            capture_input: bool = True,
            capture_metadata: bool = True,
            capture_output: bool = True,
            **attrs,
        ):
            """
            Context decorator for any arbitrary function.

            Parameters:
                - name: str | None
                    The name of the observation.
                - capture_input: bool
                    Whether to capture the input of the function.
                - capture_metadata: bool
                    Whether to capture the metadata of the function.
                - capture_output: bool
                    Whether to capture the output of the function.
                - attrs: dict
                    Additional attributes to set on the observation.

            Returns:
                A decorator function.
            """

            def decorator(fn):
                @wraps(fn)
                def observer(*args, **kwargs):
                    """
                    For both sync and async functions,
                    We need to allow applications to observe all types of contexts (e.g. generation, span, trace)

                    Parameters:
                        - args: tuple
                            The arguments to pass to the original function.
                        - kwargs: dict
                            The keyword arguments to pass to the original function.

                    Returns:
                        The result of the original function.
                    """
                    try:
                        observation_method = getattr(sdk, span_type)
                        observation_name = name or fn.__name__

                        # note: currently we only support operation type for span
                        if span_type == "span" and operation_type:
                            attrs["op"] = operation_type

                        with observation_method(
                            observation_name, **attrs
                        ) as observation:
                            if capture_input:
                                observation.input(
                                    to_json({"args": args, "kwargs": kwargs})
                                )
                            if capture_metadata and "metadata" in kwargs:
                                observation.metadata(metadata=kwargs["metadata"])

                            observation.set_attributes(**attrs)

                            try:
                                if inspect.iscoroutinefunction(fn):

                                    async def async_observer():
                                        result = await fn(*args, **kwargs)
                                        if capture_output:
                                            observation.output(to_json(result))
                                        return result

                                    return async_observer()
                                else:
                                    result = fn(*args, **kwargs)
                                    if capture_output:
                                        observation.output(to_json(result))
                                    return result
                            except Exception as e:
                                observation.error(e)
                                raise e
                    except Exception as e:
                        logging.warning(
                            "Failed to instrument function",
                            fn=fn.__name__,
                            exc_info=e,
                        )

                        return fn(*args, **kwargs)

                return observer

            return decorator

        return observe_wrapper

    sdk.observe = create_observation_decorator("span")
    sdk.observe_generation = create_observation_decorator("generation")
    sdk.observe_span = sdk.observe
    sdk.observe_tool = create_observation_decorator("span", "tool")
    sdk.observe_trace = create_observation_decorator("trace")
    return sdk
